NAME = "blue_assistant"

ICON = "🧠"

DESCRIPTION = f"{ICON} An AI Assistant."

VERSION = "4.137.1"

REPO_NAME = "blue-assistant"

MARQUEE = (
    "https://github.com/kamangir/assets/raw/main/blue-assistant/marquee.png?raw=true"
)

ALIAS = "@assistant"


def fullname() -> str:
    return f"{NAME}-{VERSION}"



from blueness import module

from blue_assistant import NAME
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


def slicing_into_frames(
    script: BaseScript,
    node_name: str,
) -> bool:
    logger.info(f"{NAME}: processing the output...")

    list_of_frame_prompts = script.nodes[node_name]["output"].split("---")
    if len(list_of_frame_prompts) != script.vars["frame_count"]:
        logger.warning(
            "{} != {}, frame count doesn't match, bad AI! 😁".format(
                len(list_of_frame_prompts),
                script.vars["frame_count"],
            )
        )

    list_of_frame_prompts += script.vars["frame_count"] * [""]

    for index in range(script.vars["frame_count"]):
        node_name = f"generating_frame_{index+1:03d}"

        script.nodes[node_name]["summary_prompt"] = list_of_frame_prompts[index]

        script.nodes[node_name]["prompt"] = (
            script.nodes[node_name]["prompt"]
            .replace(":::story", " ".join(list_of_frame_prompts[:index]))
            .replace(":::input", list_of_frame_prompts[index])
        )

    return True



from blue_objects import file, path

from blue_assistant.script.repository.generic.classes import GenericScript
from blue_assistant.script.repository.orbital_data_explorer.actions import (
    dict_of_actions,
)


class OrbitalDataExplorerScript(GenericScript):
    name = path.name(file.path(__file__))

    def perform_action(
        self,
        node_name: str,
    ) -> bool:
        if not super().perform_action(node_name=node_name):
            return False

        if node_name in dict_of_actions:
            return dict_of_actions[node_name](
                script=self,
                node_name=node_name,
            )

        return True



import copy

from blueness import module
from blue_objects import file, path

from blue_assistant import NAME
from blue_assistant.script.repository.generic.classes import GenericScript
from blue_assistant.script.repository.blue_amo.actions import dict_of_actions
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


class BlueAmoScript(GenericScript):
    name = path.name(file.path(__file__))

    def __init__(
        self,
        object_name: str,
        test_mode: bool = False,
        verbose: bool = False,
    ):
        super().__init__(
            object_name=object_name,
            test_mode=test_mode,
            verbose=verbose,
        )

        if self.test_mode:
            self.vars["frame_count"] = 1

        holder_node_name = "generating_the_frames"
        logger.info(
            "{}: expanding {} X {}...".format(
                NAME,
                holder_node_name,
                self.vars["frame_count"],
            )
        )

        holder_node = self.nodes[holder_node_name]
        del self.nodes[holder_node_name]
        self.G.remove_node(holder_node_name)

        reduce_node = "stitching_the_frames"
        self.G.add_node(reduce_node)
        self.nodes[reduce_node] = {"action": "skip"}

        for index in range(self.vars["frame_count"]):
            node_name = f"generating_frame_{index+1:03d}"

            self.nodes[node_name] = copy.deepcopy(holder_node)

            self.G.add_node(node_name)
            self.G.add_edge(
                node_name,
                "slicing_into_frames",
            )
            self.G.add_edge(
                reduce_node,
                node_name,
            )

        assert self.save_graph()

    def perform_action(
        self,
        node_name: str,
    ) -> bool:
        if not super().perform_action(node_name=node_name):
            return False

        if node_name in dict_of_actions:
            return dict_of_actions[node_name](
                script=self,
                node_name=node_name,
            )

        return True



from typing import List

from blue_options.terminal import show_usage, xtra

from blue_assistant.script.repository import list_of_script_names


def help_list(
    tokens: List[str],
    mono: bool,
) -> str:
    args = [
        "[--delim +]",
        "[--log 0]",
    ]

    return show_usage(
        [
            "@assistant",
            "script",
            "list",
        ]
        + args,
        "list scripts.",
        mono=mono,
    )


def help_run(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("download,dryrun,~upload", mono=mono)

    script_options = "script=<script>"

    args = [
        "[--test_mode 1]",
        "[--verbose 1]",
    ]

    return show_usage(
        [
            "@assistant",
            "script",
            "run",
            f"[{options}]",
            f"[{script_options}]",
            "[-|<object-name>]",
        ]
        + args,
        "run <object-name>.",
        {
            "script: {}".format(" | ".join(list_of_script_names)): [],
        },
        mono=mono,
    )


help_functions = {
    "list": help_list,
    "run": help_run,
}



from blueness import module

from blue_assistant import NAME
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


def researching_the_questions(
    script: BaseScript,
    node_name: str,
) -> bool:
    logger.info(f"{NAME}: ...")

    logger.info("🪄")

    return True



from typing import List, Type

from blue_assistant.script.repository.generic.classes import GenericScript
from blue_assistant.script.repository.blue_amo.classes import BlueAmoScript
from blue_assistant.script.repository.hue.classes import HueScript
from blue_assistant.script.repository.orbital_data_explorer.classes import (
    OrbitalDataExplorerScript,
)

list_of_script_classes: List[Type[GenericScript]] = [
    GenericScript,
    BlueAmoScript,
    HueScript,
    OrbitalDataExplorerScript,
]

list_of_script_names: List[str] = [
    script_class.name for script_class in list_of_script_classes
]



from blueness import module
from blue_objects import objects
from openai_commands.image_generation.api import OpenAIImageGenerator

from blue_assistant import NAME
from blue_assistant.env import (
    BLUE_ASSISTANT_IMAGE_DEFAULT_MODEL,
    BLUE_ASSISTANT_IMAGE_DEFAULT_SIZE,
    BLUE_ASSISTANT_IMAGE_DEFAULT_QUALITY,
)
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


# https://platform.openai.com/docs/guides/images
def generate_image(
    script: BaseScript,
    node_name: str,
) -> bool:
    logger.info(f"{NAME}: {script} @ {node_name} ...")

    generator = OpenAIImageGenerator(
        model=BLUE_ASSISTANT_IMAGE_DEFAULT_MODEL,
        verbose=script.verbose,
    )

    filename = f"{node_name}.png"

    success = generator.generate(
        prompt=script.nodes[node_name]["prompt"],
        filename=objects.path_of(
            filename=filename,
            object_name=script.object_name,
            create=True,
        ),
        quality=(BLUE_ASSISTANT_IMAGE_DEFAULT_QUALITY if script.test_mode else "hd"),
        size=(BLUE_ASSISTANT_IMAGE_DEFAULT_SIZE if script.test_mode else "1792x1024"),
        sign_with_prompt=False,
        footer=[
            script.nodes[node_name].get(
                "summary_prompt",
                script.nodes[node_name]["prompt"],
            )
        ],
    )[0]

    if success:
        script.nodes[node_name]["filename"] = filename

    return success



import os

from blue_objects import file, README

from blue_assistant import NAME, VERSION, ICON, REPO_NAME


items = [
    "[`{}`]({}) [![image]({})]({}) {}".format(
        item["title"],
        item["url"],
        item["marquee"],
        item["url"],
        item["description"],
    )
    for item in [
        {
            "title": "orbital-data-explorer",
            "url": "./blue_assistant/script/repository/orbital_data_explorer/README.md",
            "marquee": "https://github.com/kamangir/assets/raw/main/blue-plugin/marquee.png?raw=true",
            "description": "Access to the [Orbital Data Explorer](https://ode.rsl.wustl.edu/), through AI.",
        },
        {
            "title": "blue-amo",
            "url": "./blue_assistant/script/repository/blue_amo/README.md",
            "marquee": "https://github.com/kamangir/assets/raw/main/blue-amo-2025-02-03-nswnx6/stitching_the_frames-2.png?raw=true",
            "description": "A story developed and visualized, by AI.",
        },
    ]
]


def build():
    return all(
        README.build(
            items=readme.get("items", []),
            path=os.path.join(file.path(__file__), readme["path"]),
            ICON=ICON,
            NAME=NAME,
            VERSION=VERSION,
            REPO_NAME=REPO_NAME,
        )
        for readme in [
            {"items": items, "path": ".."},
            {"path": "script/repository/blue_amo"},
            {"path": "script/repository/orbital_data_explorer"},
        ]
    )



from typing import List
import numpy as np
import cv2
from tqdm import trange
import math

from blueness import module
from blue_objects import file, objects
from blue_options import string

from blue_assistant import NAME
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


def stitching_the_frames(
    script: BaseScript,
    node_name: str,
) -> bool:
    list_of_frames_filenames: List[str] = [
        filename
        for filename in [
            script.nodes[node_name_].get("filename", "")
            for node_name_ in [
                f"generating_frame_{index+1:03d}"
                for index in range(script.vars["frame_count"])
            ]
        ]
        if filename
    ]
    if not list_of_frames_filenames:
        return True

    logger.info(
        "{} frames to stitch: {}".format(
            len(list_of_frames_filenames),
            ", ".join(list_of_frames_filenames),
        )
    )

    list_of_frames: List[np.ndarray] = []
    for filename in list_of_frames_filenames:
        success, frame = file.load_image(
            objects.path_of(
                filename=filename,
                object_name=script.object_name,
            )
        )

        if success:
            list_of_frames += [frame]

    if not list_of_frames:
        return True

    common_height = list_of_frames[0].shape[0]
    common_width = list_of_frames[0].shape[1]
    for index in trange(1, len(list_of_frames)):
        frame_height = list_of_frames[index].shape[0]
        frame_width = list_of_frames[index].shape[1]

        if frame_height != common_height or frame_width != common_width:
            list_of_frames[index] = cv2.resize(
                list_of_frames[index],
                (common_width, common_height),
                interpolation=cv2.INTER_AREA,
            )

    width_count = int(math.ceil(math.sqrt(len(list_of_frames))))
    height_count = int(math.ceil(len(list_of_frames) / width_count))
    logger.info(
        "{} x {} -> {} x {}".format(
            len(list_of_frames),
            string.pretty_shape_of_matrix(list_of_frames[0]),
            height_count,
            width_count,
        )
    )

    list_of_frames += (height_count * width_count - len(list_of_frames)) * [
        np.zeros_like(list_of_frames[0])
    ]

    full_frame: np.ndarray = None
    for height_index in trange(height_count):
        row = np.concatenate(
            list_of_frames[
                height_index * width_count : (height_index + 1) * width_count
            ],
            axis=1,
        )

        if not height_index:
            full_frame = row
        else:
            full_frame = np.concatenate([full_frame, row], axis=0)
    logger.info(f"full frame: {string.pretty_shape_of_matrix(full_frame)}")

    for scale in [1, 2, 4]:
        scaled_full_frame = (
            full_frame
            if scale == 1
            else cv2.resize(
                full_frame,
                (
                    int(full_frame.shape[1] / scale),
                    int(full_frame.shape[0] / scale),
                ),
                interpolation=cv2.INTER_AREA,
            )
        )

        if not file.save_image(
            objects.path_of(
                filename="{}{}.png".format(
                    node_name,
                    "" if scale == 1 else f"-{scale}",
                ),
                object_name=script.object_name,
            ),
            scaled_full_frame,
            log=True,
        ):
            return False

    return True



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from blue_assistant import NAME
from blue_assistant.script.load import load_script
from blue_assistant.script.repository import list_of_script_names
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="list | run",
)
parser.add_argument(
    "--script_name",
    type=str,
    help=" | ".join(list_of_script_names),
)
parser.add_argument(
    "--object_name",
    type=str,
)
parser.add_argument(
    "--verbose",
    type=int,
    default=0,
    help="0 | 1",
)
parser.add_argument(
    "--test_mode",
    type=int,
    default=0,
    help="0 | 1",
)
parser.add_argument(
    "--delim",
    type=str,
    default=", ",
)
parser.add_argument(
    "--log",
    type=int,
    default=1,
    help="0 | 1",
)
args = parser.parse_args()

delim = " " if args.delim == "space" else args.delim

success = False
if args.task == "list":
    success = True
    if args.log:
        logger.info(f"{len(list_of_script_names)} script(s)")
        for index, script_name in enumerate(list_of_script_names):
            logger.info(f"#{index + 1: 3d}: {script_name}")
    else:
        print(delim.join(list_of_script_names))
elif args.task == "run":
    success, script = load_script(
        script_name=args.script_name,
        object_name=args.object_name,
        test_mode=args.test_mode == 1,
        verbose=args.verbose == 1,
    )

    if success:
        success = script.run()
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import Dict, List
import os
import networkx as nx
from functools import reduce

from blueness import module
from blue_objects import file, objects
from blue_objects.metadata import post_to_object
from blueflow.workflow import dot_file

from blue_assistant import NAME
from blue_assistant.logger import logger


NAME = module.name(__file__, NAME)


class BaseScript:
    name = "base"

    def __init__(
        self,
        object_name: str,
        test_mode: bool = False,
        verbose: bool = False,
    ):
        self.object_name = object_name

        self.test_mode = test_mode
        if self.test_mode:
            logger.info("💰 test mode is on.")

        self.verbose = verbose

        metadata_filename = os.path.join(
            file.path(__file__),
            f"../{self.name}",
            "metadata.yaml",
        )
        self.metadata: Dict
        success, self.metadata = file.load_yaml(metadata_filename)
        assert success, f"cannot load {self.name}/metadata.yaml"

        logger.info("loaded {} node(s)".format(len(self.nodes)))

        logger.info("loaded {} variable(s)".format(len(self.vars)))
        if verbose:
            for var_name, var_value in self.vars.items():
                logger.info("{}: {}".format(var_name, var_value))

        assert self.generate_graph(), "cannot generate graph"

    def apply_vars(self, text: str) -> str:
        for var_name, var_value in self.vars.items():
            text = text.replace(f":::{var_name}", str(var_value))

        return text

    def generate_graph(self) -> bool:
        self.G: nx.DiGraph = nx.DiGraph()

        list_of_nodes = list(self.nodes.keys())
        for node in self.nodes.values():
            list_of_nodes += node.get("depends-on", "").split(",")

        list_of_nodes = list({node_name for node_name in list_of_nodes if node_name})
        logger.info(
            "{} node(s): {}".format(
                len(list_of_nodes),
                ", ".join(list_of_nodes),
            )
        )

        for node_name in list_of_nodes:
            self.G.add_node(node_name)

        for node_name, node in self.nodes.items():
            for dependency in node.get("depends-on", "").split(","):
                if dependency:
                    self.G.add_edge(node_name, dependency)

        return self.save_graph()

    def get_history(
        self,
        node_name: str,
    ) -> List[str]:
        return reduce(
            lambda x, y: x + y,
            [self.get_history(successor) for successor in self.G.successors(node_name)],
            [node_name],
        )

    def run(self) -> bool:
        logger.info(
            "{}.run: {}:{} -> {}".format(
                NAME,
                self.__class__.__name__,
                self.name,
                self.object_name,
            )
        )

        return post_to_object(
            self.object_name,
            "script",
            self.script,
        )

    def save_graph(self) -> bool:
        return dot_file.save_to_file(
            objects.path_of(
                filename="workflow.dot",
                object_name=self.object_name,
            ),
            self.G,
            caption=" | ".join(
                [
                    self.name,
                    self.object_name,
                ]
            ),
            add_legend=False,
        )

    # Aliases
    @property
    def script(self) -> Dict:
        return self.metadata.get("script", {})

    @property
    def nodes(self) -> Dict[str, Dict]:
        return self.metadata.get("script", {}).get("nodes", {})

    @property
    def vars(self) -> Dict:
        return self.metadata.get("script", {}).get("vars", {})



from typing import Dict, Callable

from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.script.repository.blue_amo.actions import (
    slicing_into_frames,
    stitching_the_frames,
)


dict_of_actions: Dict[str, Callable[[BaseScript, str], bool]] = {
    "slicing_into_frames": slicing_into_frames.slicing_into_frames,
    "stitching_the_frames": stitching_the_frames.stitching_the_frames,
}



from typing import Dict, List
import os
from tqdm import tqdm


from blueness import module
from blue_objects import file, path
from blue_objects.metadata import post_to_object

from blue_assistant import NAME
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.script.actions import dict_of_actions
from blue_assistant.logger import logger


NAME = module.name(__file__, NAME)


class GenericScript(BaseScript):
    name = path.name(file.path(__file__))

    def perform_action(
        self,
        node_name: str,
    ) -> bool:
        action_name = self.nodes[node_name].get("action", "unknown")

        if action_name not in dict_of_actions:
            logger.error(f"{action_name}: action not found.")
            return False

        return dict_of_actions[action_name](
            script=self,
            node_name=node_name,
        )

    def run(
        self,
    ) -> bool:
        if not super().run():
            return False

        success: bool = True
        while (
            not all(self.nodes[node].get("completed", False) for node in self.nodes)
            and success
        ):
            for node_name in tqdm(self.nodes):
                if self.nodes[node_name].get("completed", False):
                    continue

                pending_dependencies = [
                    node_name_
                    for node_name_ in self.G.successors(node_name)
                    if not self.nodes[node_name_].get("completed", False)
                ]
                if pending_dependencies:
                    logger.info(
                        'node "{}": {} pending dependenci(es): {}'.format(
                            node_name,
                            len(pending_dependencies),
                            ", ".join(pending_dependencies),
                        )
                    )
                    continue

                if not self.perform_action(node_name=node_name):
                    success = False
                    break

                self.nodes[node_name]["completed"] = True

        if not post_to_object(
            self.object_name,
            "output",
            self.metadata,
        ):
            return False

        return success



from blue_assistant import NAME, VERSION, DESCRIPTION, REPO_NAME
from blueness.pypi import setup

setup(
    filename=__file__,
    repo_name=REPO_NAME,
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    packages=[
        NAME,
        f"{NAME}.help",
        f"{NAME}.script",
        f"{NAME}.script.actions",
        f"{NAME}.script.repository",
        f"{NAME}.script.repository.base",
        f"{NAME}.script.repository.blue_amo",
        f"{NAME}.script.repository.blue_amo.actions",
        f"{NAME}.script.repository.generic",
        f"{NAME}.script.repository.hue",
        f"{NAME}.script.repository.orbital_data_explorer",
        f"{NAME}.script.repository.orbital_data_explorer.actions",
    ],
    include_package_data=True,
    package_data={
        NAME: [
            "config.env",
            "sample.env",
            ".abcli/**/*.sh",
        ],
    },
)



from typing import List
from openai import OpenAI
import pprint

from blueness import module
from openai_commands.env import OPENAI_API_KEY

from blue_assistant import NAME
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.env import (
    BLUE_ASSISTANT_TEXT_DEFAULT_MODEL,
    BLUE_ASSISTANT_TEXT_MAX_TOKEN,
)
from blue_assistant.logger import logger

NAME = module.name(__file__, NAME)


# https://platform.openai.com/docs/guides/text-generation
def generate_text(
    script: BaseScript,
    node_name: str,
) -> bool:
    if not OPENAI_API_KEY:
        logger.error("OPENAI_API_KEY is not set.")
        return False

    logger.info(f"{NAME}: {script} @ {node_name} ...")

    messages: List = []
    list_of_context_nodes = script.get_history(node_name)
    logger.info("node context: {}".format(" <- ".join(list_of_context_nodes)))
    for context_node in reversed(list_of_context_nodes):
        messages += [
            {
                "role": "user",
                "content": [
                    {
                        "type": "text",
                        "text": script.apply_vars(script.nodes[context_node]["prompt"]),
                    }
                ],
            }
        ]

        if script.nodes[context_node].get("completed", False):
            messages += [
                {
                    "role": "assistant",
                    "content": [
                        {
                            "type": "text",
                            "text": script.nodes[context_node]["output"],
                        }
                    ],
                }
            ]

    if script.verbose:
        logger.info(f"messages: {pprint.pformat(messages)}")

    client = OpenAI(api_key=OPENAI_API_KEY)

    try:
        response = client.chat.completions.create(
            messages=messages,
            model=BLUE_ASSISTANT_TEXT_DEFAULT_MODEL,
            max_tokens=BLUE_ASSISTANT_TEXT_MAX_TOKEN,
        )
    except Exception as e:
        logger.error(str(e))
        return False

    if script.verbose:
        logger.info("response: {}".format(response))

    if not response.choices:
        logger.error("no choice.")
        return False

    output = response.choices[0].message.content
    logger.info(f"🗣️ output: {output}")
    script.nodes[node_name]["output"] = output

    return True



from typing import Dict, Callable

from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.script.repository.orbital_data_explorer.actions import (
    researching_the_questions,
)


dict_of_actions: Dict[str, Callable[[BaseScript, str], bool]] = {
    "researching_the_questions": researching_the_questions.researching_the_questions,
}



from typing import Dict, Callable, Tuple

from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.script.actions.generic import generic_action
from blue_assistant.script.actions.generate_image import generate_image
from blue_assistant.script.actions.generate_text import generate_text
from blue_assistant.script.actions.skip import skip_action
from blue_assistant.logger import logger


dict_of_actions: Dict[str, Callable[[BaseScript, str], bool]] = {
    "generic_action": generic_action,
    "generate_image": generate_image,
    "generate_text": generate_text,
    "skip": skip_action,
}



import os
from blue_options.env import load_config, load_env

load_env(__name__)
load_config(__name__)


BLUE_ASSISTANT_TEXT_DEFAULT_MODEL = os.getenv(
    "BLUE_ASSISTANT_TEXT_DEFAULT_MODEL",
    "",
)

BLUE_ASSISTANT_TEXT_MAX_TOKEN_str = os.getenv("BLUE_ASSISTANT_TEXT_MAX_TOKEN", "")
try:
    BLUE_ASSISTANT_TEXT_MAX_TOKEN = int(BLUE_ASSISTANT_TEXT_MAX_TOKEN_str)
except Exception:
    BLUE_ASSISTANT_TEXT_MAX_TOKEN = 2000


BLUE_ASSISTANT_IMAGE_DEFAULT_MODEL = os.getenv(
    "BLUE_ASSISTANT_IMAGE_DEFAULT_MODEL",
    "",
)

BLUE_ASSISTANT_IMAGE_DEFAULT_QUALITY = os.getenv(
    "BLUE_ASSISTANT_IMAGE_DEFAULT_QUALITY",
    "",
)

BLUE_ASSISTANT_IMAGE_DEFAULT_SIZE = os.getenv(
    "BLUE_ASSISTANT_IMAGE_DEFAULT_SIZE",
    "",
)



from typing import Tuple, Type

from blue_assistant.script.repository import list_of_script_classes
from blue_assistant.script.repository.base.classes import BaseScript
from blue_assistant.logger import logger


def load_script(
    script_name: str,
    object_name: str,
    test_mode: bool = False,
    verbose: bool = False,
) -> Tuple[bool, BaseScript]:
    found: bool = False
    script_class: Type[BaseScript] = BaseScript
    for script_class_option in list_of_script_classes:
        if script_class_option.name == script_name:
            found = True
            script_class = script_class_option
            break

    if not found:
        logger.error(f"{script_name}: script not found.")

    return found, script_class(
        object_name=object_name,
        test_mode=test_mode,
        verbose=verbose,
    )
